/******************************************************************************
 *  Author:		Joey Richey (CM 1698)                                         *
 *  Project:	Maze Generator                                                *
 *	Filename:	defines.h  												      *
 *                                                                            *
 *	this is a central file where I can put all my #defines and global         *
 *	functions/structures in so they will be accessible from all the files	  *
 *                                                                            *
 *****************************************************************************/
#ifndef	_DEFINES_H
#define _DEFINES_H

/*===========================
	Defines we will need
  =========================*/

/*	the first couple of defines can be changed by the user depending on the
	system	*/

/*	if you have ZThreads (http://zthread.sourceforge.net) installed and working 
	uncomment this line and the maze will be solved as a different thread while 
	the maze is displayed. If you do not have ZThreads, on linux it will try to 
	use pthreads, and on win32 you are SOL and will have it calculated before 
	the maze is displayed
*/

#define HAVE_ZTHREAD_INSTALLED

/*	for some reason, my Geforce 3 can only have so many glVertex commands
	inside glBegin/glEnd. To get around this I can do the ineffiecent thing
	of using a glBegin/glEnd for each line segment. I don't want to do this
	in general, but for my desktop it is the only way to draw a big board	*/

//#define HAVE_GHETTO_VIDEO_CARD

/*	enable/disable debug messages	*/
//#define DEBUG

/*	the percents of the screen the window should fill	*/
#define PERCENT_FILL_SCREEN		0.8

/*	the SGIs segfault when trying to get the screen width, so hardcode that
	information in	*/

#define SCREEN_WIDTH	1260
#define SCREEN_HEIGHT	980

/*	don't change anything below this line	
=============================================================================*/

/*	now that we have user defines on threads, turn those into ones my program
	can use	*/
#if (defined(_WIN32) && !defined(HAVE_ZTHREAD_INSTALLED))
	#define DONT_USE_THREADS
#endif

#if (defined(HAVE_ZTHREAD_INSTALLED))
	#define USE_ZTHREAD
#endif
#if (!defined(_WIN32) && !defined(HAVE_ZTHREAD_INSTALLED))
	#define USE_PTHREAD
#endif


/*	do includes based on what threads we need	*/
#ifdef USE_ZTHREAD
	#include <zthread/Thread.h>
	#include <zthread/FastMutex.h>

	using namespace ZThread;

	#define THREAD_CREATE(n, i)	Thread(n)
	#define THREAD_INIT(i)		
	#define	THREAD_RUN(n,i,j)	n.run( RunnablePtr(new j) )
	
	#define MUTEX_CREATE(n)	FastMutex n
	#define MUTEX_LOCK(n)	n.acquire()
	#define MUTEX_UNLOCK(n)	n.release()

#endif
#ifdef USE_PTHREAD
	#include <pthread.h>

	#define THREAD_CREATE(n, i)	pthread_t n; pthread_attr_t i
	#define THREAD_INIT(i)		pthread_attr_init(&i); \
								pthread_attr_setscope(&i,PTHREAD_SCOPE_SYSTEM);\
								pthread_attr_setschedpolicy(&i,SCHED_OTHER)
	#define THREAD_RUN(n,i,j)	pthread_create(&n, &i, j, (void *) NULL)

	#define MUTEX_CREATE(n)	pthread_mutex_t n = PTHREAD_MUTEX_INITIALIZER
	#define	MUTEX_LOCK(n)	pthread_mutex_lock(&n)
	#define MUTEX_UNLOCK(n)	pthread_mutex_unlock(&n)

#endif
#ifdef DONT_USE_THREADS
	#define THREAD_CREATE(n,i)
	#define THREAD_INIT(i)
	#define THREAD_RUN(n,i,j)

	#define MUTEX_CREATE(n)
	#define MUTEX_LOCK(n)
	#define MUTEX_UNLOCK(n)
#endif


/*	defines to make random numbers cross platform	*/
#if defined (_WIN32)
	#define SEED srand
	#define RANDNUM rand
	#define RANDMAX RAND_MAX
#else
	#define SEED srandom
	#define RANDNUM random
	#define RANDMAX (1u<<31)-1
#endif

/*	error constants	*/
#define CALLOC_MEM					1
#define RAND_LIST					2
#define UNION_CREATE				3
#define RETRIEVE_SCREEN				4
#define UNION_INIT					5		
#define UNION_MATCH_OUT_OF_BOUNDS	6
#define UNION_MERGE_OUT_OF_BOUNDS	7
#define UNSOLVEABLE_MAZE			8

/*	constants used to store for the callback lists	*/
#define MAZE			1
#define MAZE_SOLUTION	2

/*============================
	global function
============================*/

/*	error handling routine	*/
void fatalError(int);

/*============================
	Structure definitions
============================*/

/*	holds information about how big the board is	*/
struct boardInfo
{
	int width;
	int height;
} typedef boardData;

/*	holds information about the walls	*/
struct wallInfo
{
	int first;		/*	the cell grid where the wall starts			*/
	int last;		/*	the cell with the other side of the wall	*/
	bool hasWall;	/*	should a wall be drawn here?				*/
} typedef wallData;
#endif