/******************************************************************************
 *  Author:		Joey Richey (CM 1698)                                         *
 *  Project:	Maze Generator                                                *
 *	Filename:	solveMaze.cpp												  *
 *                                                                            *
 *	this file is a little hard to read with all the #ifdef, but basically     *
 *	what it boils down to is that I either make solveMaze a class or a        *
 *	function, depending on whether I use Zthreads (which requires a class).   *
 *	Either way, once it is setup, I can call it by solveMaze and by magic it  *
 *	should all work out. This class (or procedure), will solve the maze.	  *
 *                                                                            *
 *****************************************************************************/


/*===============================
	class/function solveMaze	
===============================*/
#include "solveMaze.h"

#ifndef USE_ZTHREAD
/*	need global variable if not using classes	*/
	int	**solveData;	
	std::stack<int> solve;	
#endif

#ifdef USE_ZTHREAD
void solveMaze::run(void) throw()
#else	/*	USE_ZTHREAD	*/
	/*	if i am using pthreads/no threads, i have to make solveMaze into a 
		function and not a class, so a little confusing code, but it gives a 
		different name to the function based on platform	*/
void * solveMaze(void * initWall)
#endif
	{
		/*	now that all the confusing #ifdef things are out of the way, the 
			actual function	that solves the maze begins here.	*/
		
		extern boardData board;
		extern wallData *walls;
		extern bool mazeSolved;
		extern int fastestPathLength;

		/*	TODO:HIGH:this may mess up under pthreads	*/
#ifndef DONT_USE_THREADS
		extern MUTEX_CREATE (mutex_mazeSolved);
#endif

		int i,j;

		fastestPathLength = -1;	/* we have not solved anything yet	*/

		solveData = (int **) calloc(board.width, sizeof(int *));
		if (solveData == NULL)
			fatalError(CALLOC_MEM);
		for (i=0;i<board.width;i++)
		{
			solveData[i] = (int *) calloc(board.height, sizeof(int));
			if (solveData[i] == NULL)
				fatalError(CALLOC_MEM);
			for(j=0;j<board.height;j++)
				solveData[i][j] = -1;
		}
		
		printf("Solving the maze...I bet I can do this faster than you.\n",i);

		/*	call the helper function to start solving the maze	*/
		mazeSolveHelper(0, 0, 0);	
		/*	if i never solved the maze, this is really bad and means either my
			maze data is messed up, or my solver function doesn't work	*/
		if (fastestPathLength == -1)
			fatalError(UNSOLVEABLE_MAZE);
		

		printf("I'm done trying to solve the maze and did it in %d moves\n",
				fastestPathLength);

		/*	the maze is now solved	*/
		MUTEX_LOCK(mutex_mazeSolved);
		mazeSolved=true;
		MUTEX_UNLOCK(mutex_mazeSolved);

		printf("\nPress spacebar to see my solution to the maze\n");
			
		/*	free the memory we were using	*/
		for(i=0;i<board.width;i++)
			free(solveData[i]);

		free(solveData);

#ifndef DONT_USE_THREADS
		/*	if we are using threads, we can now free the walls memory, because
			we don't need it anymore	*/
		free(walls);
#endif
#ifndef USE_ZTHREAD
		
		return 0;
#endif
	}	/*	end of solveMaze(function) or run(class)	*/

#ifdef USE_ZTHREAD
	void solveMaze::mazeSolveHelper(int x, int y, int numMoves)
#else
	void mazeSolveHelper(int x, int y, int numMoves)
#endif
	{

		/*	basic algorithim to solve a maze:
			-	check to see if the current number of moves is less than any
				previous number of moves to get to this square. if it is not,
				exit the recursion..there is a faster way to get here
			-if it is the fastest (or first):
				-	from current square check to see if i can go up. If i can
					go up, push the current location onto a stack so i can keep
					track of it. call mazeSolveHelper with new coords. when it 
					returns, pop the value off because we no loner are going 
					that way
				-	then check left, and do recursion like above
				-	then check down, and do recursion like above
				-	then check right and do recursion like above

			I know this returns a correct solution, and i _think_ it is the 
			shortest one
		*/
		extern boardData board;
		extern wallData *walls;	
		extern int* solvedPath;
		extern int fastestPathLength;
		int cellNumber;
		int i, numItems;
		std::stack<int> tmpSolve;

		/*	if this is true i have solved the maze	*/
		if ((x == board.width-1) && (y == board.height -1))
		{
			/* check to see if this is the fastest way to do it thus far	*/
			if ((numMoves < fastestPathLength) || (fastestPathLength == -1))
			{
				fastestPathLength = numMoves;
				printf("I solved the maze in %d moves...",numMoves);
				printf("perhaps there is a faster way\n");
				
				numItems = solve.size();
				/*	if we have a preivous path, free it so I don't leak 
					memeory	*/
				if (solvedPath != NULL)
					free(solvedPath);
				
				/*	create memory for the new path	*/
				solvedPath = (int*) calloc(numItems, sizeof(int));
				if (solvedPath == NULL)
					fatalError(CALLOC_MEM);

				/*	make a copy of the stack so that we can pop everythign off 
					to keep	track of how we solved it. We need to keep the 
					original around because	we may still have paths we need to 
					check	*/

				tmpSolve = solve;	/*	this seems to actually copy and not 
										just do pointers	*/
				
				for (i=0;i<numItems;i++)
				{
					solvedPath[i] = tmpSolve.top();
					tmpSolve.pop();
				}

			}
			else
			{
				printf("Maze solved again, but this time it was longer\n");
			}

		}

		/*	if I haven't been to the cell before, or this is a shorter
			route, try it	*/
		if ((numMoves < solveData[x][y]) || (solveData[x][y] == -1))
		{
			cellNumber = y * board.width + x;
			/*	try up	*/
			if(!wallExist(x,y,x,y-1))
			{
				solveData[x][y] = numMoves;
				solve.push(cellNumber);
				mazeSolveHelper(x,y-1,numMoves+1);
				solve.pop();
			}
			/*	try	left	*/
			if(!wallExist(x,y,x-1,y))
			{
				solveData[x][y] = numMoves;
				solve.push(cellNumber);
				mazeSolveHelper(x-1,y,numMoves+1);
				solve.pop();
			}
			/*	try	down	*/
			if(!wallExist(x,y,x,y+1))
			{
				solve.push(cellNumber);
				solveData[x][y] = numMoves;
				mazeSolveHelper(x,y+1,numMoves+1);
				solve.pop();
			}
			/*	try	right	*/
			if(!wallExist(x,y,x+1,y))
			{
				solve.push(cellNumber);
				solveData[x][y] = numMoves;
				mazeSolveHelper(x+1,y,numMoves+1);
				solve.pop();
			}
		}

	}

#ifdef USE_ZTHREAD
	bool solveMaze::wallExist(int x1, int y1, int x2, int y2)
#else
	bool wallExist(int x1, int y1, int x2, int y2)
#endif
	{
		extern wallData *walls;
		extern boardData board;
		int numWalls=board.height*(board.width-1)+(board.height-1)*board.width;
		int i;
		int x2Use, y2Use;

		/*	convert coords	*/
		x2Use = y1 * board.width + x1;
		y2Use = y2 * board.width + x2;

		for (i=0;i<numWalls;i++)
		{
			if (((walls[i].first == x2Use) && (walls[i].last == y2Use)) ||
			   ((walls[i].last == x2Use) && (walls[i].first == y2Use)))
			{
				return walls[i].hasWall;
			}

		}
		return true;

	}